' Fig. 23.5: FrmTicTacToeServer.vb
' This class maintains a game of Tic-Tac-Toe for two
' client applications.
Imports System.Net
Imports System.Net.Sockets
Imports System.Threading
Imports System.IO

Public Class FrmTicTacToeServer
   Private board() As Byte ' the local representation of the game board  
   Private players() As Player ' two Player objects                      
   Private playerThreads() As Thread ' Threads for client interaction    
   Private listener As TcpListener ' listen for client connection        
   Private currentPlayer As Integer ' keep track of whose turn it is         
   Private getPlayers As Thread ' Thread for acquiring client connections
   Friend disconnected As Boolean = False ' True if the server closes

   ' initialize variables and thread for receiving clients
   Private Sub FrmTicTacToeServer_Load(ByVal sender As System.Object, _
      ByVal e As System.EventArgs) Handles MyBase.Load

      board = New Byte(9) {}
      players = New Player(2) {}
      playerThreads = New Thread(2) {}
      currentPlayer = 0

      ' accept connections on a different thread         
      getPlayers = New Thread(New ThreadStart(AddressOf SetUp))
      getPlayers.Start()
   End Sub ' FrmTicTacToeServer_Load

   ' notify Players to stop Running
   Private Sub FrmTicTacToeServer_FormClosing( _
      ByVal sender As System.Object, _
      ByVal e As System.Windows.Forms.FormClosingEventArgs) _
      Handles MyBase.FormClosing

      disconnected = True
      System.Environment.Exit(System.Environment.ExitCode)
   End Sub ' FrmTicTacToeServer_FormClosing

   ' Delegate that allows method DisplayMessage to be called
   ' in the thread that creates and maintains the GUI       
   Private Delegate Sub DisplayDelegate(ByVal message As String)

   ' method DisplayMessage sets txtDisplay's Text property
   ' in a thread-safe manner
   Friend Sub DisplayMessage(ByVal message As String)
      ' if modifying txtDisplay is not thread safe
      If txtDisplay.InvokeRequired Then
         ' use inherited method Invoke to execute DisplayMessage
         ' via a Delegate                                       
         Invoke(New DisplayDelegate(AddressOf DisplayMessage), _
            New Object() {message})
         ' OK to modify txtDisplay in current thread
      Else
         txtDisplay.Text &= message
      End If
   End Sub ' DisplayMessage

   ' accepts connections from 2 players
   Public Sub SetUp()
      DisplayMessage("Waiting for players..." & vbCrLf)

      ' set up Socket                                           
      listener = New TcpListener(IPAddress.Parse("localhost"), 50000)
      listener.Start()

      ' accept first player and start a player thread              
      players(0) = New Player(listener.AcceptSocket(), Me, 0)
      playerThreads(0) = New Thread( _
         New ThreadStart(AddressOf players(0).Run))
      playerThreads(0).Start()

      ' accept second player and start another player thread       
      players(1) = New Player(listener.AcceptSocket(), Me, 1)
      playerThreads(1) = New Thread( _
         New ThreadStart(AddressOf players(1).Run))
      playerThreads(1).Start()

      ' let the first player know that the other player has connected
      SyncLock players(0)
         players(0).threadSuspended = False
         Monitor.Pulse(players(0))
      End SyncLock
   End Sub ' SetUp

   ' determine if a move is valid
   Public Function ValidMove(ByVal location As Integer, _
      ByVal player As Integer) As Boolean
      ' prevent another thread from making a move
      SyncLock Me
         ' while it is not the current player's turn, wait
         While player <> currentPlayer
            Monitor.Wait(Me)
         End While

         ' if the desired square is not occupied
         If Not IsOccupied(location) Then
            ' set the board to contain the current player's mark
            If currentPlayer = 0 Then
               board(location) = Convert.ToByte("X"c)
            Else
               board(location) = Convert.ToByte("O"c)
            End If

            ' set the currentPlayer to be the other player
            currentPlayer = (currentPlayer + 1) Mod 2

            ' notify the other player of the move                
            players(currentPlayer).OtherPlayerMoved(location)

            ' alert the other player that it's time to move
            Monitor.Pulse(Me)
            Return True
         Else
            Return False
         End If
      End SyncLock
   End Function ' ValidMove

   ' determines whether the specified square is occupied
   Public Function IsOccupied(ByVal location As Integer) As Boolean
      If board(location) = Convert.ToByte("X"c) Or _
         board(location) = Convert.ToByte("O"c) Then

         Return True
      Else
         Return False
      End If
   End Function ' IsOccupied

   ' determines if the game is over
   Public Function GameOver() As Boolean
      ' place code here to test for a winner of the game
      Return False
   End Function ' GameOver
End Class ' FrmTicTacToeServer

' class Player represents a tic-tac-toe player
Public Class Player
   Friend connection As Socket ' Socket for accepting a connection    
   Private socketStream As NetworkStream ' network data stream          
   Private server As FrmTicTacToeServer ' reference to server          
   Private writer As BinaryWriter ' facilitates writing to the stream   
   Private reader As BinaryReader ' facilitates reading from the stream 
   Private number As Integer ' player number                                
   Private mark As Char ' player's mark on the board                    
   Friend threadSuspended As Boolean = True ' if waiting for other player

   ' constructor requiring Socket, TicTacToeServerForm and Integer 
   ' objects as arguments
   Public Sub New(ByVal socket As Socket, _
      ByVal serverValue As FrmTicTacToeServer, ByVal newNumber As Integer)

      If newNumber = 0 Then
         mark = "X"c
      Else
         mark = "O"c
      End If

      connection = socket
      server = serverValue
      number = newNumber

      ' create NetworkStream object for Socket      
      socketStream = New NetworkStream(connection)

      ' create Streams for reading/writing bytes
      writer = New BinaryWriter(socketStream)
      reader = New BinaryReader(socketStream)
   End Sub ' New

   ' signal other player of move
   Public Sub OtherPlayerMoved(ByVal location As Integer)
      ' signal that opponent moved                     
      writer.Write("Opponent moved.")
      writer.Write(location) ' send location of move
   End Sub ' OtherPlayerMoved

   ' allows the players to make moves and receive moves
   ' from the other player
   Public Sub Run()
      Dim done As Boolean = False

      ' display on the server that a connection was made
      If number = 0 Then
         server.DisplayMessage("Player X connected" & vbCrLf)
      Else
         server.DisplayMessage("Player O connected" & vbCrLf)
      End If

      ' send the current player's mark to the client
      writer.Write(mark)

      ' if number equals 0 then this player is X,                
      ' otherwise O must wait for X's first move
      If number = 0 Then
         writer.Write("Player X connected.")
      Else
         writer.Write("Player O connected, please wait." & vbCrLf)
      End If

      ' X must wait for another player to arrive
      If mark = "X"c Then
         writer.Write("Waiting for another player.")

         ' wait for notification from server that another 
         ' player has connected
         SyncLock Me
            While threadSuspended
               Monitor.Wait(Me)
            End While
         End SyncLock

         writer.Write("Other player connected. Your move.")
      End If

      ' play game
      While Not done
         ' wait for data to become available
         While connection.Available = 0
            Thread.Sleep(1000)

            If server.disconnected Then
               Return
            End If
         End While

         ' receive data                   
         Dim location As Integer = reader.ReadInt32()

         ' if the move is valid, display the move on the
         ' server and signal that the move is valid
         If server.ValidMove(location, number) Then
            server.DisplayMessage("loc: " & location & vbCrLf)
            writer.Write("Valid move.")
            ' signal that the move is invalid
         Else
            writer.Write("Invalid move, try again.")
         End If

         ' if game is over, set done to True to exit while loop
         If server.GameOver() Then
            done = True
         End If
      End While

      ' close the socket connection
      writer.Close()
      reader.Close()
      socketStream.Close()
      connection.Close()
   End Sub  ' Run
End Class ' Player


'**************************************************************************
'* (C) Copyright 1992-2006 by Deitel & Associates, Inc. and               *
'* Pearson Education, Inc. All Rights Reserved.                           *
'*                                                                        *
'* DISCLAIMER: The authors and publisher of this book have used their     *
'* best efforts in preparing the book. These efforts include the          *
'* development, research, and testing of the theories and programs        *
'* to determine their effectiveness. The authors and publisher make       *
'* no warranty of any kind, expressed or implied, with regard to these    *
'* programs or to the documentation contained in these books. The authors *
'* and publisher shall not be liable in any event for incidental or       *
'* consequential damages in connection with, or arising out of, the       *
'* furnishing, performance, or use of these programs.                     *
'**************************************************************************